#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>,
# Apoorv Vyas <avyas@idiap.ch>
#

import os
import time
import unittest

import numpy as np

import torch
from torch.nn.init import normal_

from fast_transformers.aggregate import aggregate, broadcast, \
        clustered_aggregate
from fast_transformers.hashing import compute_hashes
from fast_transformers.clustering.hamming import cluster
from fast_transformers.sparse_product import sparse_dot_product
from fast_transformers.sparse_product import clustered_sparse_dot_product
from fast_transformers.masking import LengthMask


def cluster_queries(Q, query_lengths, C, I, B):
    N, H, L, E = Q.shape
    planes = Q.new_empty((B, E+1))
    normal_(planes)
    planes[:, -1] = 0
    hashes = compute_hashes(Q.view(N*H*L, E), planes).view(N, H, L)
    # Cluster the hashes and return the cluster index per query
    groups, counts = cluster(
        hashes,
        query_lengths,
        clusters=C,
        iterations=I,
        bits=B
    )

    return groups, counts


def sparse_product(Q, K, groups, topk, counts, lengths, k, Q_grouped_orig):
    N, H, L, E = Q.shape
    sorted_g, sorted_gi = torch.sort(groups.view(N*H, -1), dim=-1)
    sorted_rev_gi = torch.argsort(sorted_gi, dim=-1)

    q_offset = torch.arange(N*H, device=Q.device).unsqueeze(-1) * L
    q_flat = (sorted_gi + q_offset).reshape(-1)

    # rearrage queries
    s_queries = Q.reshape(-1, E).index_select(0, q_flat).view(N, H, L, E)
    Q_grouped = clustered_aggregate(
        s_queries, sorted_g.view(N, H, L), 1/counts.float(), lengths
    )
    topk = topk.contiguous()

    products_sorted = clustered_sparse_dot_product(
        s_queries, K, topk, groups, counts, lengths
    )
    q_rev_flat = (sorted_rev_gi + q_offset).reshape(-1)
    products = products_sorted.reshape(-1, k).index_select(0, q_rev_flat)
    products = products.view(N, H, L, k)

    return products, Q_grouped


class TestSparseProductCUDA(unittest.TestCase):
    @property
    def device(self):
        return "cuda"

    @classmethod
    def setUpClass(cls):
        if not torch.cuda.is_available():
            raise unittest.SkipTest("No CUDA capable device detected")

    def test_simple_product(self):
        N = 2
        H = 2
        L = 100
        E = 32
        S = 50
        k = 32
        C = 5
        I = 5
        B = 16

        for i in range(20):
            k = np.random.randint(10, S)
            E = np.random.randint(10, 129)
            Q = torch.randn(N, H, L, E).to(self.device)
            K = torch.randn(N, H, S, E).to(self.device)
            lengths = torch.full((N,), L, dtype=torch.int32).to(self.device)
            lengths[1] = 50
            lengths[1] = 15
            lengths[1] = 10
            groups, counts = cluster_queries(Q, lengths, C, I, B)
            Q_grouped = aggregate(Q, groups, 1/counts.float())
            QK = torch.einsum("nhle,nhse->nhls", Q_grouped, K)
            _, topk = torch.topk(QK, k, sorted=False, dim=-1)
            topk = topk.contiguous()
            products, Q_grouped_alt = sparse_product(
                Q, K, groups, topk, counts, lengths, k, Q_grouped
            )
            topk_broadcast = broadcast(
                topk.float(),
                groups,
                torch.ones_like(counts, dtype=torch.float32),
                torch.zeros((N, H, L, k), device=Q.device)
            )

            all_products = torch.einsum("nhle,nhse->nhls", Q, K)
            products_2 = all_products[
                torch.arange(N).view(N, 1, 1, 1),
                torch.arange(H).view(1, H, 1, 1),
                torch.arange(L).view(1, 1, L, 1),
                topk_broadcast.long()
            ]
            for i in range(N):
                p_1 = products[i, :, :lengths[i], :]
                p_2 = products_2[i, :, :lengths[i], :]
                self.assertLess(
                    torch.max(torch.abs(
                        p_2 - p_1
                        )
                    ),
                    1e-4
                )

    def test_difficult_product(self):
        N = 12
        H = 5
        I = 5
        B = 16

        for exp in range(30):

            C = np.random.randint(10, 500)
            L = np.random.randint(C, 2000)
            E = np.random.randint(10, 128)
            S = np.random.randint(100, 1000)
            k = np.random.randint(10, 64)

            if os.getenv("VERBOSE_TESTS", ""):
                print("Testing: N H L S E C k: {} {} {} {} {} {} {}".format(
                    N, H, L, S, E, C, k
                ))
            Q = torch.randn(N, H, L, E).to(self.device)
            K = torch.randn(N, H, S, E).to(self.device)
            lengths = torch.full((N,), L, dtype=torch.int32).to(self.device)
            groups, counts = cluster_queries(Q, lengths, C, I, B)

            Q_grouped = aggregate(Q, groups, 1/counts.float())
            QK = torch.einsum("nhle,nhse->nhls", Q_grouped, K)
            _, topk = torch.topk(QK, k, dim=-1)
            topk = topk.contiguous()

            products, _ = sparse_product(
                Q, K, groups, topk, counts, lengths, k, Q_grouped
            )
            topk_broadcast = broadcast(
                topk.float(),
                groups,
                torch.ones_like(counts, dtype=torch.float32),
                torch.zeros((N, H, L, k), device=Q.device)
            )

            all_products = torch.einsum("nhle,nhse->nhls", Q, K)
            products_2 = all_products[
                torch.arange(N).view(N, 1, 1, 1),
                torch.arange(H).view(1, H, 1, 1),
                torch.arange(L).view(1, 1, L, 1),
                topk_broadcast.long()
            ]
            self.assertLess(
                torch.max(torch.abs(
                    products_2 - products
                    )
                ),
                1e-4
            )

    @unittest.skipUnless(os.getenv("BENCHMARK_TESTS", ""), "no benchmarks")
    def test_small_benchmark(self):
        N = 12
        H = 8
        L = 1024
        E = 64
        S = 1024
        k = 32
        C = 100
        I = 10
        B = 32

        Q = torch.randn(N, H, L, E).to(self.device)
        K = torch.randn(N, H, S, E).to(self.device)
        lengths = torch.full((N,), L, dtype=torch.int32).to(self.device)
        groups, counts = cluster_queries(Q, lengths, C, I, B)

        sorted_g, sorted_gi = torch.sort(groups.view(N*H, -1), dim=-1)
        sorted_rev_gi = torch.argsort(sorted_gi, dim=-1)

        q_offset = torch.arange(N*H, device=Q.device).unsqueeze(-1) * L
        q_flat = (sorted_gi + q_offset).reshape(-1)

        # sorted queries, keys, values
        s_queries = Q.reshape(-1, E).index_select(0, q_flat).view(N, H, L, E)
        Q_grouped = aggregate(
            s_queries, sorted_g.view(N, H, L), 1/counts.float()
        )
        QK = torch.einsum("nhle,nhse->nhls", Q_grouped, K)
        _, topk = torch.topk(QK, k, dim=-1)
        topk = topk.contiguous()

        products_sorted = clustered_sparse_dot_product(
            s_queries,
            K,
            topk,
            groups,
            counts,
            lengths
        )
        q_rev_flat = (sorted_rev_gi + q_offset).reshape(-1)
        products = products_sorted.reshape(-1, k).index_select(0, q_rev_flat)
        products = products.view(N, H, L, k)

        for i in range(1000):
            products_sorted = clustered_sparse_dot_product(
                s_queries,
                K,
                topk,
                groups,
                counts,
                lengths
            )

        torch.cuda.synchronize()
        s = torch.cuda.Event(enable_timing=True)
        e = torch.cuda.Event(enable_timing=True)
        s.record()
        products_sorted = clustered_sparse_dot_product(
            s_queries,
            K,
            topk,
            groups,
            counts,
            lengths
        )
        e.record()
        torch.cuda.synchronize()
        t_sc = s.elapsed_time(e)

        products_sorted = products_sorted.reshape(-1, k).index_select(0, q_rev_flat).view(N, H, L, k)
        topk = topk.contiguous()
        print("Sparse_Clustered: {}".format(t_sc))


if __name__ == "__main__":
    unittest.main()



#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>
#

"""Define composable functions to filter events."""

import weakref

from .event import Event


class EventFilter(object):
    """EventFilter instances are predicates (ie functions that return True or
    False) to be used with an event dispatcher for filtering event
    instances.

    The main benefit from using raw functions is that an EventFilter composes
    very easily using operators such as &, |, ~.

    Example
    --------

        event_filter = AttentionEvent | layer_name_contains("layers.1")
        event_filter = from_layer(transformer.layers[2].attention)
        event_filter = (
            AttentionEvent &
            lambda ev: torch.isnan(ev.attention_matrix).any()
        )
    """
    def __call__(self, event):
        raise NotImplementedError()

    def _to_event_filter(self, other):
        if isinstance(other, EventFilter):
            return other
        if isinstance(other, type) and issubclass(other, Event):
            return event_class(other)
        if callable(other):
            return CallableEventFilter(other)

        return NotImplemented

    def __and__(self, other):
        other = self._to_event_filter(other)
        if other is NotImplemented:
            return other
        return CallableEventFilter(lambda ev: self(ev) and other(ev))

    def __rand__(self, other):
        other = self._to_event_filter(other)
        if other is NotImplemented:
            return other
        return CallableEventFilter(lambda ev: other(ev) and self(ev))

    def __or__(self, other):
        other = self._to_event_filter(other)
        if other is NotImplemented:
            return other
        return CallableEventFilter(lambda ev: self(ev) or other(ev))

    def __ror__(self, other):
        other = self._to_event_filter(other)
        if other is NotImplemented:
            return other
        return CallableEventFilter(lambda ev: other(ev) or self(ev))

    def __invert__(self):
        return CallableEventFilter(lambda ev: not self(ev))


class CallableEventFilter(EventFilter):
    """Wrap a function with an EventFilter object."""
    def __init__(self, event_filter):
        self._event_filter = event_filter

    def __call__(self, event):
        return self._event_filter(event)


class LayerNameEventFilter(EventFilter):
    """A LayerNameEventFilter allows to filter events based on a human readable
    name of the layer that emitted them.

    Note that LayerNameEventFilter keeps a weak reference to all modules which
    means that it cannot be used to prevent modules from being garbage
    collected.

    Arguments
    ---------
        root: torch.nn.Module instance that represents the root container
        name_filter: callable, that returns true if the name 
    """
    def __init__(self, root, name_filter):
        self._names = {
            weakref.ref(m): n
            for n, m in root.named_modules()
        }
        self._name_filter = name_filter

    def __call__(self, event):
        name = self._names.get(weakref.ref(event.source), None)
        if name is None:
            return False
        return self._name_filter(name)


def event_class(klass):
    """Select events that are instances of `klass`.

    Arguments
    ---------
        klass: A class to check the event instance against

    Returns
    -------
        An instance of EventFilter
    """
    return CallableEventFilter(lambda ev: isinstance(ev, klass))


def from_layer(layer):
    """Select events that are dispatched from the `layer`.

    Arguments
    ---------
        layer: An instance of torch.nn.Module to check against the event source

    Returns
    -------
        An instance of EventFilter
    """
    return CallableEventFilter(lambda ev: ev.source is layer)


def layer_name_contains(root, name):
    """Select events that contain `name` in their human readable name.

    We use root.named_modules() to get human readable names for the layers.
    """
    return LayerNameEventFilter(root, lambda n: name in n)



#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>
#

import unittest

from fast_transformers.events import Event, EventDispatcher


class MockEvent(Event):
    def __init__(self, source, payload):
        super(MockEvent, self).__init__(source)
        self.payload = payload


class TestEventDispatcher(unittest.TestCase):
    def test_simple_listen_dispatch(self):
        d = {"x": 0}
        def listener1(event):
            d["x"] += 1

        def listener2(event):
            d["x"] += 1

        ed = EventDispatcher()
        ed.listen(Event, listener1)
        ed.listen(Event, listener2)
        ed.dispatch(Event(None))
        self.assertEqual(d["x"], 2)
        ed.remove(listener1)
        ed.dispatch(Event(None))
        self.assertEqual(d["x"], 3)
        ed.remove(listener2)

        def set_payload(event):
            d.update(event.payload)
        ed.listen(MockEvent, set_payload)
        ed.dispatch(Event(None))
        self.assertTrue("y" not in d)
        ed.dispatch(MockEvent(None, {"y": 0}))
        self.assertEqual(d["y"], 0)
        self.assertEqual(d["x"], 3)

    def test_factory_method(self):
        ed1 = EventDispatcher.get()
        ed2 = EventDispatcher.get()
        self.assertTrue(ed1 is ed2)
        ed1 = EventDispatcher.get("foo")
        ed2 = EventDispatcher.get("bar")
        self.assertTrue(ed1 is not ed2)
        ed1 = EventDispatcher.get("foo")
        ed2 = EventDispatcher.get("foo")
        self.assertTrue(ed1 is ed2)
        ed1 = EventDispatcher()
        ed2 = EventDispatcher.get(ed1)
        self.assertTrue(ed1 is ed2)



if __name__ == "__main__":
    unittest.main()



#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>
#

import unittest

import torch

from fast_transformers.events import EventDispatcher, QKVEvent, AttentionEvent
from fast_transformers.events.filters import layer_name_contains
from fast_transformers.builders import TransformerEncoderBuilder


class TestEvents(unittest.TestCase):
    def test_qkv(self):
        d = {}
        def store_qkv(event):
            d["q"] = event.queries
            d["k"] = event.keys
            d["v"] = event.values
        # default transformer is 4 layers 4 heads
        transformer = TransformerEncoderBuilder().get()
        x = transformer(torch.rand(1, 100, 64*4))
        self.assertEqual(len(d), 0)

        EventDispatcher.get().listen(QKVEvent, store_qkv)
        x = transformer(torch.rand(1, 100, 64*4))
        self.assertEqual(len(d), 3)
        d.clear()

        EventDispatcher.get().remove(store_qkv)
        x = transformer(torch.rand(1, 100, 64*4))
        self.assertEqual(len(d), 0)
        d.clear()

        EventDispatcher.get().listen(
            QKVEvent & layer_name_contains(transformer, "layers.2.attention"),
            store_qkv
        )
        x = transformer(torch.rand(1, 100, 64*4))
        self.assertEqual(len(d), 3)
        d.clear()

        EventDispatcher.get().listen(
            QKVEvent & layer_name_contains(transformer, "layers.22.attention"),
            store_qkv
        )
        x = transformer(torch.rand(1, 100, 64*4))
        self.assertEqual(len(d), 0)
        d.clear()
        EventDispatcher.get().clear()

    def test_attention_matrix(self):
        A = []
        def store_attention(event):
            A.append(event.attention_matrix)
        # default transformer is 4 layers 4 heads
        transformer = TransformerEncoderBuilder().get()
        x = transformer(torch.rand(1, 100, 64*4))
        self.assertEqual(len(A), 0)

        EventDispatcher.get().listen(AttentionEvent, store_attention)
        x = transformer(torch.rand(1, 100, 64*4))
        self.assertEqual(len(A), 4)



if __name__ == "__main__":
    unittest.main()



#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>,
# Apoorv Vyas <avyas@idiap.ch>
#


import unittest
import os
import numpy as np
import time

import torch

try:
    from fast_transformers.aggregate import clustered_aggregate, \
            clustered_broadcast
except ImportError:
    pass


class TestAggregateCPU(unittest.TestCase):

    def test_aggregate(self):
        N = 2
        H = 4
        L = 80
        E = 2
        C = 4

        for i in range(30):
            C = np.random.randint(5, 10)
            L = np.random.randint(1, 30) * C
            E = np.random.randint(10, 128)
            if os.getenv("VERBOSE_TESTS", ""):
                print(("Testing: N H L E C: "
                       "{} {} {} {} {}").format(N, H, L, E, C))

            x = torch.rand((N, H, L, E)).cpu()
            g = (torch.arange(L) % C).view(1, 1, L).repeat(N, H, 1).int().cpu()
            f = torch.ones(N, H, C).cpu() * (C / L)
            counts = torch.ones_like(f, dtype=torch.int32) * (L // C)
            y = torch.zeros(N, H, C, E).cpu()
            lengths = torch.full((N,), L, dtype=torch.int32).to(x.device)

            sorted_g, sorted_gi = torch.sort(g.view(N*H, -1), dim=-1)
            sorted_rev_gi = torch.argsort(sorted_gi, dim=-1)

            q_offset = torch.arange(N*H, device=x.device).unsqueeze(-1) * L
            q_flat = (sorted_gi + q_offset).reshape(-1)

            # sorted queries, keys, values
            s_x = x.reshape(-1, E).index_select(0, q_flat).view(N, H, L, E)
            y = clustered_aggregate(
                s_x, sorted_g.view(N, H, -1), f, lengths, y
            )
            for i in range(C):
                self.assertLess(
                    torch.abs(
                        x[:, :, i::C, :].mean(2) - y[:, :, i, :]
                    ).max().item(),
                    1e-6
                )

    def test_aggregate_masked(self):
        N = 10
        H = 3
        L = 40
        E = 32
        C = 4

        for i in range(30):
            C = np.random.randint(5, 10)
            L = np.random.randint(2, 30) * C
            E = np.random.randint(10, 128)
            if os.getenv("VERBOSE_TESTS", ""):
                print(("Testing: N H L E C: "
                       "{} {} {} {} {}").format(N, H, L, E, C))

            x = torch.rand((N, H, L, E)).cpu()
            g = (torch.arange(L) % C).view(1, 1, L).repeat(N, H, 1).int().cpu()
            g[:, :, -C:] = C + 1
            c = (L // C) - 1

            lengths = torch.full((N,), L-C, dtype=torch.int32).to(x.device)
            f = torch.ones(N, H, C).cpu() / float(c)
            counts = torch.ones_like(f, dtype=torch.int32) * c
            y = torch.zeros(N, H, C, E).cpu()

            sorted_g, sorted_gi = torch.sort(g.view(N*H, -1), dim=-1)
            sorted_rev_gi = torch.argsort(sorted_gi, dim=-1)

            q_offset = torch.arange(N*H, device=x.device).unsqueeze(-1) * L
            q_flat = (sorted_gi + q_offset).reshape(-1)

            # sorted queries, keys, values
            s_x = x.reshape(-1, E).index_select(0, q_flat).view(N, H, L, E)
            y = clustered_aggregate(
                s_x, sorted_g.view(N, H, -1), f, lengths, y
            )

            for i in range(C):
                x_m = x[:, :, i::C, :][:, :, :-1, :].mean(2)
                self.assertLess(
                    torch.abs(
                        x_m - y[:, :, i, :]
                    ).max().item(),
                    1e-6
                )


if __name__ == "__main__":
    unittest.main()



#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>,
# Apoorv Vyas <avyas@idiap.ch>
#

import os
from os import getenv
import unittest
import numpy as np

import torch
from torch.nn.init import normal_

from fast_transformers.aggregate import aggregate, broadcast
from fast_transformers.hashing import compute_hashes
from fast_transformers.clustering.hamming import cluster
from fast_transformers.sparse_product import clustered_sparse_dot_product
from fast_transformers.masking import LengthMask


def cluster_queries(Q, query_lengths, C, I, B):
    N, H, L, E = Q.shape
    planes = Q.new_empty((B, E+1))
    normal_(planes)
    planes[:, -1] = 0
    hashes = compute_hashes(Q.view(N*H*L, E), planes).view(N, H, L)
    # Cluster the hashes and return the cluster index per query
    groups, counts = cluster(
        hashes,
        query_lengths,
        clusters=C,
        iterations=I,
        bits=B
    )

    return groups, counts


def sparse_product(Q, K, groups, topk, counts, lengths):
    N, H, L, E = Q.shape
    _, _, C, k = topk.shape

    sorted_g, sorted_gi = torch.sort(groups.view(N*H, -1), dim=-1)
    sorted_rev_gi = torch.argsort(sorted_gi, dim=-1)

    q_offset = torch.arange(N*H, device=Q.device).unsqueeze(-1) * L
    q_flat = (sorted_gi + q_offset).reshape(-1)

    # sorted queries, keys, values
    s_queries = Q.reshape(-1, E).index_select(0, q_flat).view(N, H, L, E)
    Q_grouped = aggregate(s_queries, sorted_g.view(N, H, L), 1/counts.float())
    topk = topk.contiguous()

    products_sorted = clustered_sparse_dot_product(
        s_queries, K, topk, sorted_g.view(N, H, L), counts, lengths
    )
    q_rev_flat = (sorted_rev_gi + q_offset).reshape(-1)
    products = products_sorted.reshape(-1, k).index_select(
        0, q_rev_flat).view(N, H, L, k)
    return products


class TestSparseProductBackward(unittest.TestCase):
    @property
    def device(self):
        return "cpu"

    def _zero_grad(self, Q, K):
        for x in [Q, K]:
            if x.grad is not None:
                x.grad[...] = 0

    def test_masked_simple_grad(self):
        N = 4
        H = 2
        L = 100
        E = 64
        S = 100
        k = 32
        C = 5
        I = 5
        B = 16

        for i in range(30):
            C = np.random.randint(10, 500)
            L = np.random.randint(C, 2000)
            E = np.random.randint(10, 128)
            S = np.random.randint(100, 1000)
            k = np.random.randint(10, 64)

            if os.getenv("VERBOSE_TESTS", ""):
                print(("Testing Masked: N H L S E C k: "
                       "{} {} {} {} {} {} {}").format(N, H, L, S, E, C, k))

            Q = torch.randn(N, H, L, E).to(self.device).requires_grad_(True)
            K = torch.randn(N, H, S, E).to(self.device).requires_grad_(True)

            lengths = np.random.randint(C, L+1, N)
            lengths = torch.tensor(lengths, dtype=torch.int32).to(self.device)
            query_lengths = LengthMask(
                lengths,
                max_len=L
            )
            groups, counts = cluster_queries(Q, lengths, C, I, B)
            Q_grouped = aggregate(Q, groups, 1/counts.float())
            QK = torch.einsum("nhle,nhse->nhls", Q_grouped, K)
            _, topk = torch.topk(QK, k, dim=-1)
            topk = topk.contiguous()
            topk_broadcast = broadcast(
                topk.float(),
                groups,
                torch.ones_like(counts, dtype=torch.float32),
                torch.zeros((N, H, L, k), device=Q.device)
            )

            self._zero_grad(Q, K)
            QK_full = torch.einsum("nhle,nhse->nhls", Q, K)
            QK_selected = QK_full[
                torch.arange(N).view(N, 1, 1, 1).to(self.device),
                torch.arange(H).view(1, H, 1, 1).to(self.device),
                torch.arange(L).view(1, 1, L, 1).to(self.device),
                topk_broadcast.long()
            ]

            QK_selected = QK_selected * query_lengths.float_matrix[:, None, :, None]
            QK_selected.sum().backward()
            grad = [torch.clone(Q.grad), torch.clone(K.grad)]

            self._zero_grad(Q, K)
            QK_selected_hat = sparse_product(
                Q, K, groups, topk, counts, lengths
            )
            QK_selected_hat.sum().backward()
            grad_hat = [torch.clone(Q.grad), torch.clone(K.grad)]
            self.assertLess(
                torch.abs(QK_selected - QK_selected_hat).max(),
                1e-4
            )
            for g1, g2 in zip(grad, grad_hat):
                self.assertLess(
                    torch.abs(g1 - g2).max(),
                    1e-3
                )

    def test_simple_grad(self):
        N = 2
        H = 2
        L = 100
        E = 64
        S = 100
        k = 32
        C = 5
        I = 5
        B = 16

        Q = torch.randn(N, H, L, E).to(self.device).requires_grad_(True)
        K = torch.randn(N, H, S, E).to(self.device).requires_grad_(True)

        lengths = torch.full((N,), L, dtype=torch.int32).to(self.device)
        groups, counts = cluster_queries(Q, lengths, C, I, B)
        Q_grouped = aggregate(Q, groups, 1/counts.float())
        QK = torch.einsum("nhle,nhse->nhls", Q_grouped, K)
        _, topk = torch.topk(QK, k, dim=-1)
        topk = topk.contiguous()
        topk_broadcast = broadcast(
            topk.float(),
            groups,
            torch.ones_like(counts, dtype=torch.float32),
            torch.zeros((N, H, L, k), device=Q.device)
        )

        self._zero_grad(Q, K)
        QK_full = torch.einsum("nhle,nhse->nhls", Q, K)
        QK_selected = QK_full[
            torch.arange(N).view(N, 1, 1, 1).to(self.device),
            torch.arange(H).view(1, H, 1, 1).to(self.device),
            torch.arange(L).view(1, 1, L, 1).to(self.device),
            topk_broadcast.long()
        ]

        QK_selected.sum().backward()
        grad = [torch.clone(Q.grad), torch.clone(K.grad)]

        self._zero_grad(Q, K)
        QK_selected_hat = sparse_product(Q, K, groups, topk, counts, lengths)
        QK_selected_hat.sum().backward()
        grad_hat = [torch.clone(Q.grad), torch.clone(K.grad)]
        self.assertLess(
            torch.abs(QK_selected - QK_selected_hat).max(),
            1e-4
        )
        for g1, g2 in zip(grad, grad_hat):
            self.assertLess(
                torch.abs(g1 - g2).max(),
                1e-4
            )

    def test_difficult_grad(self):
        N = 12
        H = 5
        I = 5
        B = 16

        for exp in range(30):
            C = np.random.randint(10, 500)
            L = np.random.randint(C, 2000)
            E = np.random.randint(10, 128)
            S = np.random.randint(100, 1000)
            k = np.random.randint(10, 64)

            if os.getenv("VERBOSE_TESTS", ""):
                print(("Testing: N H L S E C k: "
                       "{} {} {} {} {} {} {}").format(N, H, L, S, E, C, k))

            Q = torch.randn(N, H, L, E).to(self.device)
            K = torch.randn(N, H, S, E).to(self.device)
            Q.requires_grad = True
            K.requires_grad = True
            lengths = torch.full((N,), L, dtype=torch.int32).to(self.device)
            groups, counts = cluster_queries(Q, lengths, C, I, B)
            Q_grouped = aggregate(Q, groups, 1/counts.float())
            QK = torch.einsum("nhle,nhse->nhls", Q_grouped, K)
            _, topk = torch.topk(QK, k, dim=-1)
            topk = topk.contiguous()
            topk_broadcast = broadcast(
                topk.float(),
                groups,
                torch.ones_like(counts, dtype=torch.float32),
                torch.zeros((N, H, L, k), device=Q.device)
            )

            self._zero_grad(Q, K)
            QK_full = torch.einsum("nhle,nhse->nhls", Q, K)
            QK_selected = QK_full[
                torch.arange(N).view(N, 1, 1, 1).to(self.device),
                torch.arange(H).view(1, H, 1, 1).to(self.device),
                torch.arange(L).view(1, 1, L, 1).to(self.device),
                topk_broadcast.long()
            ]

            QK_selected.sum().backward()
            grad = [torch.clone(Q.grad), torch.clone(K.grad)]

            self._zero_grad(Q, K)
            QK_selected_hat = sparse_product(
                Q, K, groups, topk, counts, lengths
            )
            QK_selected_hat.sum().backward()
            grad_hat = [torch.clone(Q.grad), torch.clone(K.grad)]

            self.assertLess(
                torch.abs(QK_selected - QK_selected_hat).max(),
                1e-4
            )
            i = 0
            for g1, g2 in zip(grad, grad_hat):
                self.assertLess(
                    torch.abs(g1 - g2).max(),
                    1e-3
                )
                i += 1


if __name__ == "__main__":
    unittest.main()



#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>
#

import unittest

import torch

from fast_transformers.attention import AttentionLayer, LinearAttention
from fast_transformers.feature_maps.fourier_features import \
    RandomFourierFeatures, SmoothedRandomFourierFeatures, Favor, \
    GeneralizedRandomFeatures
from fast_transformers.masking import FullMask


class TestFourierFeatures(unittest.TestCase):
    def test_omega(self):
        f = RandomFourierFeatures(32, n_dims=64, orthogonal=True)
        f.new_feature_map("cpu")
        self.assertLess(
            torch.abs(
                f.omega.t().matmul(f.omega)[torch.eye(32) == 0]
            ).max().item(),
            1e-4
        )

    def test_rff(self):
        for ortho in [False, True]:
            f = RandomFourierFeatures(32, n_dims=32*1000, softmax_temp=1,
                                      orthogonal=ortho)
            f.new_feature_map("cpu")

            x = torch.randn(100, 32) * 0.15
            y = torch.randn(100, 32) * 0.15
            phi_x = f(x)
            phi_y = f(y)

            rbf_xy = torch.exp(-((x[:, None] - y[None, :])**2).sum(-1)/2)
            rbf_xy_hat = phi_x.matmul(phi_y.t())

            self.assertLess(
                ((rbf_xy - rbf_xy_hat)**2).mean().item(),
                1e-4
            )

            f = SmoothedRandomFourierFeatures(32, n_dims=32*1000,
                                              softmax_temp=1, orthogonal=ortho,
                                              smoothing=1.0)
            f.new_feature_map("cpu")
            phi_x = f(x)
            phi_y = f(y)
            rbf_xy = torch.exp(-((x[:, None] - y[None, :])**2).sum(-1)/2) + 1
            rbf_xy_hat = phi_x.matmul(phi_y.t())

            self.assertLess(
                ((rbf_xy - rbf_xy_hat)**2).mean().item(),
                1e-4
            )

    def test_prf(self):
        for ortho in [False, True]:
            f = Favor(32, n_dims=32*1000, softmax_temp=1, orthogonal=ortho)

            f.new_feature_map("cpu")

            x = torch.randn(100, 32) * 0.15
            y = torch.randn(100, 32) * 0.15
            phi_x = f(x)
            phi_y = f(y)

            sm_xy = torch.exp(x.mm(y.t()))
            sm_xy_hat = phi_x.mm(phi_y.t())

            self.assertLess(
                ((sm_xy - sm_xy_hat)**2).mean().item(),
                1e-3
            )

    def test_grf(self):
        f = GeneralizedRandomFeatures(32, n_dims=128)
        f.new_feature_map("cpu")
        x = torch.randn(100, 32)
        phi_x = f(x)
        self.assertEqual((100, 128), phi_x.shape)

    def test_feature_map_sharing(self):
        x = torch.rand(3, 100, 4*32)
        f = Favor.factory(n_dims=64)
        att = AttentionLayer(
            LinearAttention(32, f),
            4*32,
            4
        )

        attn_mask = FullMask(100)
        lengths = FullMask(3, 100)
        y = att(x, x, x, attn_mask, lengths, lengths)
        y = att(y, y, y, attn_mask, lengths, lengths)
        y.sum().backward()

    def test_feature_map_redraw(self):
        x = torch.rand(3, 100, 32)
        f = Favor(32, n_dims=64)

        f.new_feature_map("cpu")
        fx1 = f(x)
        f.new_feature_map("cpu")
        fx2 = f(x)
        self.assertFalse(torch.allclose(fx1, fx2))

        f = Favor(32, n_dims=64, redraw=2)
        f.new_feature_map("cpu")
        fx1 = f(x)
        f.new_feature_map("cpu")
        fx2 = f(x)
        f.new_feature_map("cpu")
        fx3 = f(x)
        self.assertTrue(torch.allclose(fx1, fx2))
        self.assertFalse(torch.allclose(fx2, fx3))

        f = Favor(32, n_dims=64, redraw=1, deterministic_eval=True)
        f.new_feature_map("cpu")
        fx1 = f(x)
        f.new_feature_map("cpu")
        fx2 = f(x)
        f.eval()
        f.new_feature_map("cpu")
        fx3 = f(x)
        self.assertFalse(torch.allclose(fx1, fx2))
        self.assertTrue(torch.allclose(fx2, fx3))


if __name__ == "__main__":
    unittest.main()



#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>
#

import unittest

import torch

from fast_transformers.events import Event
from fast_transformers.events.filters import event_class, from_layer, \
    layer_name_contains


class MockEvent(Event):
    def __init__(self, source, payload):
        super(MockEvent, self).__init__(source)
        self.payload = payload


class TestEventFilters(unittest.TestCase):
    def test_simple_filters(self):
        mock_event = event_class(MockEvent)
        self.assertTrue(mock_event(MockEvent(None, None)))
        self.assertFalse(mock_event(Event(None)))

        source = object()
        fl = from_layer(source)
        self.assertTrue(fl(Event(source)))
        self.assertFalse(fl(Event(None)))
        self.assertFalse(fl(Event(object())))

        net = torch.nn.Sequential(
            torch.nn.Linear(2, 10),
            torch.nn.ReLU(),
            torch.nn.Sequential(
                torch.nn.Linear(10, 10),
                torch.nn.Linear(10, 10),
                torch.nn.Linear(10, 10)
            ),
            torch.nn.ReLU(),
            torch.nn.Linear(10, 1),
            torch.nn.Sigmoid()
        )
        lnc = layer_name_contains(net, "2.1")
        self.assertTrue(lnc(Event(net[2][1])))

    def test_filter_composition(self):
        net = torch.nn.Sequential(
            torch.nn.Linear(2, 10),
            torch.nn.ReLU(),
            torch.nn.Sequential(
                torch.nn.Linear(10, 10),
                torch.nn.Linear(10, 10),
                torch.nn.Linear(10, 10)
            ),
            torch.nn.ReLU(),
            torch.nn.Linear(10, 1),
            torch.nn.Sigmoid()
        )

        event_filter = MockEvent & from_layer(net[2][1])
        self.assertFalse(event_filter(Event(net[2][1])))
        self.assertFalse(event_filter(MockEvent(net[2], None)))
        self.assertTrue(event_filter(MockEvent(net[2][1], None)))

        # should raise error because ev.payload is accessed before event is
        # made sure to be a MockEvent
        event_filter = (lambda ev: ev.payload == 0) & event_class(MockEvent)
        with self.assertRaises(AttributeError):
            event_filter(Event(None))
        event_filter = event_class(MockEvent) & (lambda ev: ev.payload == 0)
        self.assertFalse(event_filter(Event(None)))
        self.assertFalse(event_filter(MockEvent(None, 1)))
        self.assertTrue(event_filter(MockEvent(None, 0)))


if __name__ == "__main__":
    unittest.main()



#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>,
# Apoorv Vyas <avyas@idiap.ch>
#


import unittest
import os
import numpy as np
import time

import torch
from torch.nn.init import normal_
from torch.nn import functional as F

from fast_transformers.hashing import compute_hashes
from fast_transformers.clustering.hamming import cluster
from fast_transformers.aggregate import aggregate, broadcast
from fast_transformers.aggregate import clustered_broadcast, \
        clustered_aggregate


def cluster_queries(Q, query_lengths, C, I, B):
    N, H, L, E = Q.shape
    planes = Q.new_empty((B, E+1))
    normal_(planes)
    planes[:, -1] = 0
    hashes = compute_hashes(Q.view(N*H*L, E), planes).view(N, H, L)
    # Cluster the hashes and return the cluster index per query
    groups, counts = cluster(
        hashes,
        query_lengths,
        clusters=C,
        iterations=I,
        bits=B
    )

    return groups, counts


def try_sorted_broadcast(Q, K, V, groups, counts, lengths, Q_grouped_orig):
    N, H, L, E = Q.shape
    _, _, S, D = V.shape
    sorted_g, sorted_gi = torch.sort(groups.view(N*H, -1), dim=-1)
    sorted_rev_gi = torch.argsort(sorted_gi, dim=-1)

    q_offset = torch.arange(N*H, device=Q.device).unsqueeze(-1) * L
    q_flat = (sorted_gi + q_offset).reshape(-1)

    # sorted queries, keys, values
    s_queries = Q.reshape(-1, E).index_select(0, q_flat).view(N, H, L, E)
    Q_grouped = clustered_aggregate(
        s_queries, sorted_g.view(N, H, L), 1 / counts.float(), lengths
    )
    assert(abs(Q_grouped_orig - Q_grouped).max().item() < 1e-4)
    QK = torch.einsum("nhle,nhse->nhls", Q_grouped, K)
    A = F.softmax(QK, dim=-1)
    V_new = torch.einsum("nhls,nhse->nhle", A, V)
    V_broadcast = torch.zeros((N, H, L, E), dtype=V_new.dtype).cpu()
    factors = torch.ones_like(counts, dtype=torch.float)
    V_sorted_broadcast = clustered_broadcast(
        V_new, sorted_g.view(N, H, L), counts, factors, V_broadcast
    )
    q_rev_flat = (sorted_rev_gi + q_offset).reshape(-1)
    V_broadcast_remap = V_sorted_broadcast.reshape(-1, D).index_select(
        0, q_rev_flat).view(N, H, L, D)
    return V_broadcast_remap


class TestClusteredBroadcastCPU(unittest.TestCase):

    def test_broadcast_full(self):
        N = 3
        H = 2
        L = 400
        S = 100
        E = 256
        C = 211
        I = 5
        B = 16

        for exp in range(50):
            Q = torch.randn(N, H, L, E).cpu()
            lengths = torch.full((N,), L, dtype=torch.int32).cpu()
            lengths[0] = np.random.randint(C, L+1)
            lengths[1] = np.random.randint(C, L+1)
            groups, counts = cluster_queries(Q, lengths, C, I, B)
            Q_grouped = aggregate(Q, groups, 1/counts.float())
            K = torch.randn(N, H, S, E).cpu()
            QK = torch.einsum("nhle,nhse->nhls", Q_grouped, K)

            V = torch.randn(N, H, S, E).cpu()
            A = F.softmax(QK, dim=-1)
            V_new = torch.einsum("nhls,nhse->nhle", A, V)
            V_broadcast_2 = broadcast(
                V_new,
                groups,
                torch.ones_like(counts, dtype=torch.float32),
                torch.zeros((N, H, L, E), device=Q.device)
            )
            V_broadcast = try_sorted_broadcast(
                Q, K, V, groups, counts, lengths, Q_grouped
            )
            self.assertLess(
                torch.max(torch.abs(
                    V_broadcast_2
                    - V_broadcast
                    )
                ),
                1e-4
            )

    def test_broadcast_difficult(self):
        N = 10
        H = 3
        E = 64
        I = 5
        B = 16

        for exp in range(20):
            S = np.random.randint(100, 1000)
            C = np.random.randint(10, 500)
            L = np.random.randint(C, 2000)
            E = np.random.randint(10, 160)
            lengths = torch.tensor(
                np.random.randint(C, L+1, N),
                dtype=torch.int32
            ).cpu()
            if os.getenv("VERBOSE_TESTS", ""):
                print(("Test: N H L S E C: "
                       "{} {} {} {} {} {}").format(N, H, L, S, E, C))

            Q = torch.randn(N, H, L, E).cpu()
            groups, counts = cluster_queries(Q, lengths, C, I, B)
            Q_grouped = aggregate(Q, groups, 1/counts.float())
            K = torch.randn(N, H, S, E).cpu()
            QK = torch.einsum("nhle,nhse->nhls", Q_grouped, K)

            V = torch.randn(N, H, S, E).cpu()
            A = F.softmax(QK, dim=-1)
            V_new = torch.einsum("nhls,nhse->nhle", A, V)
            V_broadcast_2 = broadcast(
                V_new,
                groups,
                torch.ones_like(counts, dtype=torch.float32),
                torch.zeros((N, H, L, E), device=Q.device)
            )
            V_broadcast = try_sorted_broadcast(
                Q, K, V, groups, counts, lengths, Q_grouped
            )
            self.assertLess(
                torch.max(torch.abs(
                    V_broadcast_2
                    - V_broadcast
                    )
                ),
                1e-4
            )


if __name__ == "__main__":
    unittest.main()



#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>,
# Apoorv Vyas <avyas@idiap.ch>
#

"""Similar to the corresponding module in fast_transformers.attention, this
module performs all the query, key, value projections and output projections
leaving the implementation of the attention to the inner attention module."""

from torch.nn import Linear, Module

from ....events import EventDispatcher
from ..._utils import check_state


class RecurrentAttentionLayer(Module):
    """See fast_transformers.attention.attention_layer.AttentionLayer.

    The only difference with the corresponding module is that this projects
    only one input and then calls the inner attention with the provided
    previous state.

    Arguments
    ---------
        attention: Specific inner attention implementation that just computes a
                   weighted average of values given a similarity of queries and
                   keys.
        d_model: The input feature dimensionality
        n_heads: The number of heads for the multi head attention
        d_keys: The dimensionality of the keys/queries
                (default: d_model/n_heads)
        d_values: The dimensionality of the values (default: d_model/n_heads)
        event_dispatcher: str or EventDispatcher instance to be used by this
                          module for dispatching events (default: the default
                          global dispatcher)
    """
    def __init__(self, attention, d_model, n_heads, d_keys=None,
                 d_values=None, event_dispatcher=""):
        super(RecurrentAttentionLayer, self).__init__()

        # Fill d_keys and d_values
        d_keys = d_keys or (d_model//n_heads)
        d_values = d_values or (d_model//n_heads)

        self.inner_attention = attention
        self.query_projection = Linear(d_model, d_keys * n_heads)
        self.key_projection = Linear(d_model, d_keys * n_heads)
        self.value_projection = Linear(d_model, d_values * n_heads)
        self.out_projection = Linear(d_values * n_heads, d_model)
        self.n_heads = n_heads
        self.event_dispatcher = EventDispatcher.get(event_dispatcher)

    def forward(self, query, key, value, state=None, memory=None):
        """Apply attention to the passed in query/key/value after projecting
        them to multiple heads.

        In the argument description we make use of the following sizes

            - N: the batch size
            - D: The input feature dimensionality passed in the constructor as
              'd_model'

        Arguments
        ---------
            query: (N, D) The tensor containing the queries
            key: (N, D) The tensor containing the keys
            value: (N, D) The tensor containing the values
            state: The state varies depending on the inner attention implementation
            memory: **Deprecated** and replaced by state

        Returns
        -------
            The new value for each query as a tensor of shape (N, D).
        """
        # Normalize the state/memory
        state = check_state(state, memory)

        # Project the queries/keys/values
        query = self.query_projection(query)
        key = self.key_projection(key)
        value = self.value_projection(value)

        # Reshape them into many heads and compute the attention
        N, D = query.shape
        H = self.n_heads
        new_value, state = self.inner_attention(
            query.view(N, H, -1),
            key.view(N, H, -1),
            value.view(N, H, -1),
            state
        )
        new_value = new_value.view(N, -1)

        # Project the output and return
        return self.out_projection(new_value), state



#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>,
# Apoorv Vyas <avyas@idiap.ch>
#

import os
from os import getenv
import unittest

import torch

from fast_transformers.sparse_product import sparse_weighted_average


class TestSparseWeightedAverage(unittest.TestCase):
    @property
    def device(self):
        return "cuda"

    @classmethod
    def setUpClass(cls):
        if not torch.cuda.is_available():
            raise unittest.SkipTest("No CUDA capable device detected")

    def _zero_grad(self, Q, K):
        for x in [Q, K]:
            if x.grad is not None:
                x.grad[...] = 0

    def test_correctness(self):
        N = 2
        H = 4
        L = 3000
        S = 3000
        E = 32
        k = 32
        weights = torch.rand(N, H, L, k).to(self.device).requires_grad_(True)
        values = torch.randn(N, H, S, E).to(self.device).requires_grad_(True)

        attn = torch.randn(N, H, L, S).to(self.device).requires_grad_(False) 
        topk_v, topk = torch.topk(attn, k, dim=-1)

        self._zero_grad(weights, values)
        values_selected = values[
            torch.arange(N).view(N, 1, 1, 1).to(self.device),
            torch.arange(H).view(1, H, 1, 1).to(self.device),
            topk
        ]
        output = (weights.unsqueeze(-1)*values_selected).sum(-2)
        output.sum().backward()
        grad = [torch.clone(weights.grad), torch.clone(values.grad)]

        self._zero_grad(weights, values)
        output_hat = sparse_weighted_average(weights, values, topk)
        output_hat.sum().backward()
        grad_hat = [torch.clone(weights.grad), torch.clone(values.grad)]
        self.assertLess(
            torch.abs(output - output_hat).max(),
            1e-4
        )
        for g1, g2 in zip(grad, grad_hat):
            self.assertLess(
                torch.abs(g1 - g2).max(),
                1e-4
            )

    def test_forward(self):
        N = 5
        H = 2 
        L = 100
        S = 100
        E = 32
        k = 5

        weights = torch.arange(0,k).expand(N, H, L, k).to(self.device).float().requires_grad_(True)
        values = torch.arange(0,E).expand(N, H, L, E).to(self.device).float().requires_grad_(True)

        attn = torch.arange(0, S).expand(N, H, L, S).to(self.device).float().requires_grad_(False)
        topk_v, topk = torch.topk(attn, k, dim=-1)

        values_selected = values[
            torch.arange(N).view(N, 1, 1, 1).to(self.device),
            torch.arange(H).view(1, H, 1, 1).to(self.device),
            topk
        ]
        output = (weights.unsqueeze(-1)*values_selected).sum(-2)
        output_hat = sparse_weighted_average(weights, values, topk)
        self.assertLess(
            torch.abs(output - output_hat).max(),
            1e-4
        )

    @unittest.skipUnless(os.getenv("BENCHMARK_TESTS", ""), "no benchmarks")
    def test_benchmark_forward(self):
        N = 12
        H = 8
        L = 2000
        S = 2000
        E = 32
        k = 32

        weights = torch.rand(N, H, L, k).to(self.device).requires_grad_(True)
        values = torch.randn(N, H, S, E).to(self.device).requires_grad_(True)
        
        attn = torch.randn(N, H, L, S).to(self.device).requires_grad_(False) 
        topk_v, topk = torch.topk(attn, k, dim=-1)
        topk = topk.contiguous()
        for i in range(2000):
            output_hat = sparse_weighted_average(weights, values, topk)

        s = torch.cuda.Event(enable_timing=True)
        e = torch.cuda.Event(enable_timing=True)
        s.record()
        output_hat = sparse_weighted_average(weights, values, topk)
        e.record()
        torch.cuda.synchronize()
        t_sparse = s.elapsed_time(e)

        print('T_sparse Forward:{}'.format(t_sparse))

    @unittest.skipUnless(os.getenv("BENCHMARK_TESTS", ""), "no benchmarks")
    def test_benchmark_backward(self):
        N = 12
        H = 8
        L = 2000
        S = 2000
        E = 32
        k = 32

        weights = torch.rand(N, H, L, k).to(self.device).requires_grad_(True)
        values = torch.randn(N, H, S, E).to(self.device).requires_grad_(True)

        attn = torch.randn(N, H, L, S).to(self.device).requires_grad_(False)
        topk_v, topk = torch.topk(attn, k, dim=-1)
        topk = topk.contiguous()
        for i in range(2000):
            output_hat = sparse_weighted_average(weights, values, topk)

        self._zero_grad(weights, values)
        s = torch.cuda.Event(enable_timing=True)
        e = torch.cuda.Event(enable_timing=True)
        s.record()
        output_hat.sum().backward()
        e.record()
        torch.cuda.synchronize()
        t_sparse = s.elapsed_time(e)

        print('T_sparse Backward:{}'.format(t_sparse))

if __name__ == "__main__":
    unittest.main()



#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>
#

"""Implement the positive orthogonal random features from the paper
"Rethinking Attention with Performers" https://arxiv.org/pdf/2009.14794.pdf
and the traditional random Fourier features that approximate the RBF kernel.
"""

from math import sqrt, log
import warnings

import torch

from .base import FeatureMap


def orthogonal_random_matrix_(w):
    """Initialize the matrix w in-place to compute orthogonal random features.

    The matrix is initialized such that its columns are orthogonal to each
    other (in groups of size `rows`) and their norms is drawn from the
    chi-square distribution with `rows` degrees of freedom (namely the norm of
    a `rows`-dimensional vector distributed as N(0, I)).

    Arguments
    ---------
        w: float tensor of size (rows, columns)
    """
    rows, columns = w.shape
    start = 0
    while start < columns:
        end = min(start+rows, columns)
        block = torch.randn(rows, rows, device=w.device)
        norms = torch.sqrt(torch.einsum("ab,ab->a", block, block))
        Q, _ = torch.qr(block)
        w[:, start:end] = (
            Q[:, :end-start] * norms[None, :end-start]
        )
        start += rows


class RandomFourierFeatures(FeatureMap):
    """Random Fourier Features for the RBF kernel according to [1].

    [1]: "Weighted Sums of Random Kitchen Sinks: Replacing minimization with
         randomization in learning" by A. Rahimi and Benjamin Recht.

    Arguments
    ---------
        query_dimensions: int, The input query dimensions in order to sample
                          the noise matrix
        n_dims: int, The size of the feature map (should be divisible by 2)
                (default: query_dimensions)
        softmax_temp: float, The temerature for the Gaussian kernel
                      approximation exp(-t * |x-y|^2)
                      (default: 1/sqrt(query_dimensions))
        orthogonal: bool, When True the random matrix is initialized for
                    orthogonal random features to reduce the approximation
                    variance (default: False)
        redraw: int, Redraw the random matrix every 'redraw' times
                (default: 1)
        deterministic_eval: bool, Only redraw the random matrix during training
                            (default: False)
    """
    def __init__(self, query_dimensions, n_dims=None, softmax_temp=None,
                 orthogonal=False, redraw=1, deterministic_eval=False):
        super(RandomFourierFeatures, self).__init__(query_dimensions)

        self.n_dims = n_dims or query_dimensions
        self.query_dimensions = query_dimensions
        self.orthogonal = orthogonal
        self.softmax_temp = (
            1/sqrt(query_dimensions) if softmax_temp is None
            else softmax_temp
        )
        self.redraw = redraw
        self.deterministic_eval = deterministic_eval

        # Make a buffer for storing the sampled omega
        self.register_buffer(
            "omega",
            torch.zeros(self.query_dimensions, self.n_dims//2)
        )
        self._calls = -1

    def new_feature_map(self, device):
        # If we are not training skip the generation of a new feature map
        if self.deterministic_eval and not self.training:
            return

        # Only redraw the new feature map every self.redraw times
        self._calls += 1
        if (self._calls % self.redraw) != 0:
            return

        omega = torch.zeros(
            self.query_dimensions,
            self.n_dims//2,
            device=device
        )
        if self.orthogonal:
            orthogonal_random_matrix_(omega)
        else:
            omega.normal_()
        self.register_buffer("omega", omega)

    def forward(self, x):
        x = x * sqrt(self.softmax_temp)
        u = x.unsqueeze(-2).matmul(self.omega).squeeze(-2)
        phi = torch.cat([torch.cos(u), torch.sin(u)], dim=-1)
        return phi * sqrt(2/self.n_dims)


class SmoothedRandomFourierFeatures(RandomFourierFeatures):
    """Simply add a constant value to the dot product in order to avoid
    possible numerical instabilities when the feature map is slightly
    negative.

    Implements K(x, y) = exp(-|x-y|^2) + s.

    Arguments
    ---------
        query_dimensions: int, The input query dimensions in order to sample
                          the noise matrix
        n_dims: int, The size of the feature map (should be divisible by 2)
                (default: query_dimensions)
        softmax_temp: float, The temerature for the Gaussian kernel
                      approximation exp(-t * |x-y|^2)
                      (default: 1/sqrt(query_dimensions))
        orthogonal: bool, When True the random matrix is initialized for
                    orthogonal random features to reduce the approximation
                    variance (default: False)
        smoothing: float, The smoothing parameter to add to the dot product.
        redraw: int, Redraw the random matrix every 'redraw' times
                (default: 1)
        deterministic_eval: bool, Only redraw the random matrix during training
                            (default: False)
    """
    def __init__(self, query_dimensions, n_dims=None, softmax_temp=None,
                 orthogonal=False, smoothing=1.0, redraw=1,
                 deterministic_eval=False):
        super(SmoothedRandomFourierFeatures, self).__init__(
            query_dimensions,
            n_dims=query_dimensions-1 if n_dims is None else n_dims-1,
            softmax_temp=softmax_temp,
            orthogonal=orthogonal,
            redraw=redraw,
            deterministic_eval=deterministic_eval
        )
        self.smoothing = smoothing

    def forward(self, x):
        y = super().forward(x)
        smoothing = torch.full(
            y.shape[:-1] + (1,),
            self.smoothing,
            dtype=y.dtype,
            device=y.device
        )
        return torch.cat([y, smoothing], dim=-1)


class Favor(RandomFourierFeatures):
    """Positive orthogonal random features that approximate the softmax kernel.

    Basically implementation of Lemma 1 from "Rethinking Attention with
    Performers".

    Arguments
    ---------
        query_dimensions: int, The input query dimensions in order to sample
                          the noise matrix
        n_dims: int, The size of the feature map (should be divisible by 2)
                (default: query_dimensions)
        softmax_temp: float, The temerature for the softmax approximation
                     (default: 1/sqrt(query_dimensions))
        orthogonal: bool, If set to true then the random matrix should be
                    orthogonal which results in lower approximation variance
                    (default: True)
        stabilize: bool, If set to True subtract the max norm from the
                   exponentials to make sure that there are no infinities. It
                   is equivalent to a robust implementation of softmax where
                   the max is subtracted before the exponentiation.
                   (default: False)
        redraw: int, Redraw the random matrix every 'redraw' times
                (default: 1)
        deterministic_eval: bool, Only redraw the random matrix during training
                            (default: False)
    """
    def __init__(self, query_dimensions, n_dims=None, softmax_temp=None,
                 orthogonal=True, stabilize=False, redraw=1,
                 deterministic_eval=False):
        super(Favor, self).__init__(query_dimensions, n_dims=n_dims,
                                    softmax_temp=softmax_temp,
                                    orthogonal=orthogonal, redraw=redraw,
                                    deterministic_eval=deterministic_eval)
        self.stabilize = stabilize

    def _check_sequence_length(self, x):
        """Check that the 2nd dimension is larger than the 3rd as a heuristic
        that the sequence length will be larger than the number of heads. If
        not simply warn of a possible bug."""
        if len(x.shape) != 4:
            warnings.warn(("Favor.stabilize is set to True but the input "
                           "feature does not have the shape (N, L, H, D) "
                           "which may result in unexpected behaviour"))

        if x.shape[1] < x.shape[2]:
            warnings.warn(("Favor.stabilize is set to True but the 2nd "
                           "dimension of the input is smaller than the 3rd "
                           "which could indicate that the sequence length and "
                           "the heads are flipped. This may result in incorrect "
                           "behaviour. The shape of the input is "
                           "{!r}.").format(x.shape))

    def forward(self, x):
        x = x * sqrt(self.softmax_temp)
        norm_x_squared = torch.einsum("...d,...d->...", x, x).unsqueeze(-1)
        u = x.unsqueeze(-2).matmul(self.omega).squeeze(-2)

        # Compute the offset for the exponential such that h(x) is multiplied
        # in logspace. In particular, we multiply with exp(-norm_x_squared/2)
        # and 1/sqrt(self.n_dims)
        offset = norm_x_squared * 0.5 + 0.5 * log(self.n_dims)

        # If stabilize is True then add the max norm per sequence in order to
        # ensure that exp_u1 and exp_u2 will be <1.
        #
        # NOTE: This is the only part of this feature map that assumes the
        #       2nd dimension is the sequence length. We call the
        #       _check_sequence_length dimension function to be able to catch
        #       some possible bugs ahead of time.
        if self.stabilize:
            self._check_sequence_length(norm_x_squared)
            offset = offset + norm_x_squared.max(1, keepdim=True)[0]

        exp_u1 = torch.exp(u - offset)
        exp_u2 = torch.exp(-u - offset)
        phi = torch.cat([exp_u1, exp_u2], dim=-1)

        return phi


class GeneralizedRandomFeatures(RandomFourierFeatures):
    """Implements the generalized random Fourier features from Performers.

    It computes φ(χ) = [f(ω_1 χ), f(ω_2 χ), ..., f(ω_n χ)] where f(.) is the
    passed in `kernel_fn`.

    Arguments
    ---------
        query_dimensions: int, The input query dimensions in order to sample
                          the noise matrix
        n_dims: int, The size of the feature map (default: query_dimensions)
        softmax_temp: float, A normalizer for the dot products that is
                     multiplied to the input features before the feature map
                     application (default: 1.0)
        orthogonal: bool, If set to true then the random matrix should be
                    orthogonal which results in lower approximation variance
                    (default: True)
        kernel_fn: callable, defines the f used for the feature map.
                   (default: relu)
        redraw: int, Redraw the random matrix every 'redraw' times
                (default: 1)
        deterministic_eval: bool, Only redraw the random matrix during training
                            (default: False)
    """
    def __init__(self, query_dimensions, n_dims=None, softmax_temp=1.0,
                 orthogonal=True, kernel_fn=torch.relu, redraw=1,
                 deterministic_eval=False):
        super(GeneralizedRandomFeatures, self).__init__(
            query_dimensions,
            n_dims=2*query_dimensions if n_dims is None else 2*n_dims,
            softmax_temp=softmax_temp,
            orthogonal=orthogonal,
            redraw=redraw,
            deterministic_eval=deterministic_eval
        )
        self.kernel_fn = kernel_fn

    def forward(self, x):
        if self.softmax_temp != 1.0:
            x = x * sqrt(self.softmax_temp)
        u = x.unsqueeze(-2).matmul(self.omega).squeeze(-2)
        return self.kernel_fn(u)



#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>,
# Apoorv Vyas <avyas@idiap.ch>
#


import torch

from .aggregate_cpu import aggregate as aggregate_cpu, \
    broadcast as broadcast_cpu
try:
    from .aggregate_cuda import aggregate as aggregate_gpu, \
        broadcast as broadcast_gpu
    from .clustered_aggregate_cuda import \
        clustered_broadcast as clustered_broadcast_gpu, \
        clustered_aggregate as clustered_aggregate_gpu

except ImportError:
    pass


def aggregate(X, G, F, Y=None):
    device = X.device
    if Y is None:
        Y = torch.zeros(
            F.shape + (X.shape[-1],),
            device=device,
            dtype=X.dtype
        )
    else:
        Y.zero_()

    if device.type == "cpu":
        aggregate_cpu(X, G, F, Y)
    else:
        aggregate_gpu(X, G, F, Y)

    return Y


def broadcast(Y, G, F, X=None):
    device = Y.device
    if X is None:
        X = torch.zeros(
            G.shape + (Y.shape[-1],),
            device=device,
            dtype=Y.dtype
        )

    if device.type == "cpu":
        broadcast_cpu(Y, G, F, X)
    else:
        broadcast_gpu(Y, G, F, X)

    return X


# Divide the cluster into groups of equal size
# as constrained by the shared memory
def set_group(C, E):
    C_per_block = int(192 * 64 / (E+1))
    G_min = (C + C_per_block - 1) // C_per_block
    for G in range(G_min, C+1):
        if C % G == 0:
            return G


def clustered_broadcast(Y, groups, counts, factors, X=None):
    device = Y.device
    if X is None:
        X = torch.zeros(
            groups.shape + (Y.shape[-1],),
            device=device,
            dtype=Y.dtype
        )
    if device.type == "cpu":
        broadcast_cpu(Y, groups, factors, X)
    else:
        N, H, C, E = Y.shape
        _, _, L, _ = X.shape

        # Following are some booking keeping parameters to facilitate the
        # broadcast kernel that takes advantage of clustering
        # More information can be found in the cuda file
        with torch.no_grad():
            threads = 256
            G = set_group(C, E)
            group_counts = counts.view(N, H, G, -1).sum(-1)
            block_counts = (group_counts + threads - 1) // threads
            total_blocks = block_counts.sum().item()
            indx_maps = torch.ones(
                (total_blocks, 5),
                device=X.device,
                dtype=torch.int32
            )

        clustered_broadcast_gpu(
            Y,
            groups,
            factors,
            X,
            block_counts.int(),
            group_counts.int(),
            threads,
            G,
            total_blocks,
            indx_maps
        )
    return X


def clustered_aggregate(X, G, F, lengths, Y=None):
    device = X.device
    if Y is None:
        Y = torch.zeros(
            F.shape + (X.shape[-1],),
            device=device,
            dtype=X.dtype
        )
    else:
        Y.zero_()

    if device.type == "cpu":
        aggregate_cpu(X, G, F, Y)
    else:
        clustered_aggregate_gpu(X, G, F, lengths, Y)
    return Y



#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>
#

"""This module implements a basic event system that allows the transformer
internal components to make available any tensor with minimal overhead."""

from .event import Event, AttentionEvent, QKVEvent
from .event_dispatcher import EventDispatcher



#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>
#

"""Create the feature map interface and some commonly used feature maps.

All attention implementations that expect a feature map shall receive a factory
function that returns a feature map instance when called with the query
dimensions.
"""

from functools import partial

import torch
from torch.nn import Module


class FeatureMap(Module):
    """Define the FeatureMap interface."""
    def __init__(self, query_dims):
        super().__init__()
        self.query_dims = query_dims

    def new_feature_map(self, device):
        """Create a new instance of this feature map. In particular, if it is a
        random feature map sample new parameters."""
        raise NotImplementedError()

    def forward_queries(self, x):
        """Encode the queries `x` using this feature map."""
        return self(x)

    def forward_keys(self, x):
        """Encode the keys `x` using this feature map."""
        return self(x)

    def forward(self, x):
        """Encode x using this feature map. For symmetric feature maps it
        suffices to define this function, but for asymmetric feature maps one
        needs to define the `forward_queries` and `forward_keys` functions."""
        raise NotImplementedError()

    @classmethod
    def factory(cls, *args, **kwargs):
        """Return a function that when called with the query dimensions returns
        an instance of this feature map.

        It is inherited by the subclasses so it is available in all feature
        maps.
        """
        def inner(query_dims):
            return cls(query_dims, *args, **kwargs)
        return inner


class ActivationFunctionFeatureMap(FeatureMap):
    """Define a feature map that is simply an element-wise activation
    function."""
    def __init__(self, query_dims, activation_function):
        super().__init__(query_dims)
        self.activation_function = activation_function

    def new_feature_map(self, device):
        return

    def forward(self, x):
        return self.activation_function(x)


elu_feature_map = ActivationFunctionFeatureMap.factory(
    lambda x: torch.nn.functional.elu(x) + 1
)



#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>
#

"""Implement local context attention."""

from math import sqrt

import torch
from torch.nn import Module, Dropout
from torch.nn import functional as F

from ..attention_registry import AttentionRegistry, Optional, Int, Float, \
    EventDispatcherInstance
from ..events import EventDispatcher
from ..local_product import local_dot_product, local_weighted_average


class LocalAttention(Module):
    """Implement fast local attention where a query can only attend to
    neighboring keys.

    In this attention module the query Q_i can only attend to a key K_j if
    |i-j| < local_context/2.

    Arguments
    ---------
        local_context: The neighborhood to consider for local attention.
        softmax_temp: The temperature to use for the softmax attention.
                      (default: 1/sqrt(d_keys) where d_keys is computed at
                      runtime)
        attention_dropout: The dropout rate to apply to the attention
                           (default: 0.1)
        event_dispatcher: str or EventDispatcher instance to be used by this
                          module for dispatching events (default: the default
                          global dispatcher)
    """
    def __init__(self, local_context, softmax_temp=None, attention_dropout=0.1,
                 event_dispatcher=""):
        super(LocalAttention, self).__init__()
        self.local_context = local_context
        self.softmax_temp = softmax_temp
        self.dropout = Dropout(attention_dropout)
        self.event_dispatcher = EventDispatcher.get(event_dispatcher)

    def forward(self, queries, keys, values, attn_mask, query_lengths,
                key_lengths):
        """Implements the local attention.

        The attn_mask can be anything but the only values that will be
        considered will be the ones in the neighborhood of each query.

        Arguments
        ---------
            queries: (N, L, H, E) The tensor containing the queries
            keys: (N, S, H, E) The tensor containing the keys
            values: (N, S, H, D) The tensor containing the values
            attn_mask: An implementation of BaseMask that encodes where each
                       query can attend to
            query_lengths: An implementation of  BaseMask that encodes how
                           many queries each sequence in the batch consists of
            key_lengths: An implementation of BaseMask that encodes how
                         many queries each sequence in the batch consists of
        """
        # Extract some shapes and compute the temperature
        N, L, H, E = queries.shape
        _, S, _, D = values.shape
        context = self.local_context
        softmax_temp = self.softmax_temp or 1./sqrt(E)

        # Permute the dimensions to NHLE instead of NLHE
        queries = queries.permute(0, 2, 1, 3).contiguous()
        keys = keys.permute(0, 2, 1, 3).contiguous()
        values = values.permute(0, 2, 1, 3).contiguous()

        QK = local_dot_product(
            queries,
            keys,
            attn_mask.additive_matrix_finite,
            key_lengths.lengths,
            self.local_context
        )
        A = self.dropout(torch.softmax(softmax_temp * QK, dim=-1))

        V_new = local_weighted_average(A, values)

        return V_new.permute(0, 2, 1, 3).contiguous()


# Register the attention implementation so that it becomes available in our
# builders
AttentionRegistry.register(
    "local", LocalAttention,
    [
        ("local_context", Int),
        ("softmax_temp", Optional(Float)),
        ("attention_dropout", Optional(Float, 0.1)),
        ("event_dispatcher", Optional(EventDispatcherInstance, ""))
    ]
)



#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>
#

"""Implementations of feature maps to be used with linear attention and causal
linear attention."""


from .base import elu_feature_map, ActivationFunctionFeatureMap
from .fourier_features import RandomFourierFeatures, Favor, \
    SmoothedRandomFourierFeatures, GeneralizedRandomFeatures



#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>
#

from collections import OrderedDict

from .event import Event
from .filters import event_class


class EventDispatcher(object):
    """An EventDispatcher is a simple way to implement an observer pattern for
    loose coupling of components. In our case it is used so that the internals
    of large neural networks can communicate with the outside world in an
    agnostic and efficient way.

    Example usage
    -------------

        from fast_transformers.events import EventDispatcher, AttentionEvent
        from fast_transformers.events.filters import \
            layer_name_contains

        def attention_event_handler(event):
            print(event.attention_matrix)

        ed = EventDispatcher()
        ed.listen(AttentionEvent, attention_event_handler)
        ed.listen(
            AttentionEvent & layer_name_contains("layers.12"),
            attention_event_handler
        )
    """
    _dispatchers = {}

    def __init__(self):
        self._listeners = OrderedDict()

    def listen(self, event_filter, event_handler):
        """Add an event handler for the events that pass the event filter.

        Arguments
        ---------
            event_filter: callable or Event class to define for which events
                          this handler will be called
            event_handler: callable that accepts an instance of Event
        """
        if isinstance(event_filter, type) and issubclass(event_filter, Event):
            event_filter = event_class(event_filter)

        self._listeners[event_handler] = event_filter

    def remove(self, event_handler):
        """Remove the event_handler from the listeners so that no more events
        are dispatched to this handler."""
        self._listeners.pop(event_handler, None)

    def clear(self):
        """Remove all listeners from the event dispatcher."""
        self._listeners.clear()

    def dispatch(self, event):
        """Dispatch an event to the listeners.

        Arguments
        ---------
            event: Event instance
        """
        for event_handler, event_filter in self._listeners.items():
            if event_filter(event):
                event_handler(event)

    @classmethod
    def get(cls, key=""):
        """Factory method for creating global event dispatchers for loosely
        coupling parts of a larger codebase.

        Since global objects are a complete antipattern, we suggest that this
        is only used to set a default value for an event dispatcher passed as
        an argument.

        Argument
        --------
            key: A key to uniquely identify a dispatcher or an instance of a
                 dispatcher to be returned as is
        """
        if isinstance(key, cls):
            return key
        if key not in cls._dispatchers:
            cls._dispatchers[key] = cls()
        return cls._dispatchers[key]



#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>,
# Apoorv Vyas <avyas@idiap.ch>
#

import os
from os import getenv
import unittest

import numpy as np

import torch
from torch.nn.init import normal_

from fast_transformers.aggregate import aggregate, broadcast
from fast_transformers.hashing import compute_hashes
from fast_transformers.clustering.hamming import cluster
from fast_transformers.sparse_product import clustered_sparse_weighted_average
from fast_transformers.sparse_product import clustered_sparse_dot_product
from fast_transformers.masking import LengthMask


def cluster_queries(Q, query_lengths, C, I, B):
    N, H, L, E = Q.shape
    planes = Q.new_empty((B, E+1))
    normal_(planes)
    planes[:, -1] = 0
    hashes = compute_hashes(Q.view(N*H*L, E), planes).view(N, H, L)
    # Cluster the hashes and return the cluster index per query
    groups, counts = cluster(
        hashes,
        query_lengths,
        clusters=C,
        iterations=I,
        bits=B
    )

    return groups, counts


class TestSparseWeightedAverage(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        if not torch.cuda.is_available():
            raise unittest.SkipTest("No CUDA capable device detected")

    @property
    def device(self):
        return "cuda"

    def _zero_grad(self, Q, K):
        for x in [Q, K]:
            if x.grad is not None:
                x.grad[...] = 0

    def test_correctness_masked(self):
        N = 12
        H = 6
        L = 1000
        S = 1000
        E = 32
        k = 32
        C = 100
        I = 10
        B = 32
        for exp in range(30):
            N = np.random.randint(1, 6)
            H = np.random.randint(1, 8)
            C = np.random.randint(10, 500)
            L = np.random.randint(C, 2000)
            E = np.random.randint(10, 128)
            S = np.random.randint(100, 1000)
            k = np.random.randint(10, 64)

            if os.getenv("VERBOSE_TESTS", ""):
                print(("Testing Masked: N H L S E C k: "
                       "{} {} {} {} {} {} {}").format(N, H, L, S, E, C, k))

            Q = torch.randn(N, H, L, E).to(self.device)
            K = torch.randn(N, H, S, E).to(self.device)

            lengths = np.random.randint(C, L+1, N)
            lengths = torch.tensor(lengths, dtype=torch.int32).to(self.device)
            lengths[0] = L
            query_lengths = LengthMask(
                lengths,
                max_len=L
            )
            groups, counts = cluster_queries(Q, lengths, C, I, B)

            sorted_g, sorted_gi = torch.sort(groups.view(N*H, -1), dim=-1)
            sorted_rev_gi = torch.argsort(sorted_gi, dim=-1)

            q_offset = torch.arange(N*H, device=Q.device).unsqueeze(-1) * L
            q_flat = (sorted_gi + q_offset).reshape(-1)
            s_queries = Q.reshape(-1, E).index_select(
                0, q_flat).view(N, H, L, E)
            Q_grouped = aggregate(
                s_queries, sorted_g.view(N, H, L), 1/counts.float()
            )

            QK = torch.einsum("nhle,nhse->nhls", Q_grouped, K)
            _, topk = torch.topk(QK, k, dim=-1)
            topk = topk.contiguous()
            topk_broadcast = broadcast(
                topk.float(),
                groups,
                torch.ones_like(counts, dtype=torch.float32),
                torch.zeros((N, H, L, k), device=Q.device)
            )
            weights_sorted = torch.rand(
                N, H, L, k).to(self.device).requires_grad_(True)
            weights_sorted.retain_grad()

            q_rev_flat = (sorted_rev_gi + q_offset).reshape(-1)
            weights = torch.clone(
                weights_sorted.reshape(-1, k).index_select(
                    0, q_rev_flat
                ).view(N, H, L, k)
            )
            weights.retain_grad()

            values = torch.randn(
                N, H, S, E).to(self.device).requires_grad_(True)
            self._zero_grad(weights, values)
            values_selected = values[
                torch.arange(N).view(N, 1, 1, 1).to(self.device),
                torch.arange(H).view(1, H, 1, 1).to(self.device),
                topk_broadcast.long()
            ]
            output = (weights.unsqueeze(-1)*values_selected).sum(-2)
            output = output * query_lengths.float_matrix[:, None, :, None]
            output.sum().backward()
            grad = [torch.clone(weights.grad), torch.clone(values.grad)]

            self._zero_grad(weights_sorted, values)
            self._zero_grad(weights, values)

            output_hat_sorted = clustered_sparse_weighted_average(
                weights_sorted, values, topk, groups, counts
            )
            output_hat = output_hat_sorted.reshape(
                -1, E).index_select(0, q_rev_flat).view(N, H, L, E)

            self.assertLess(
                torch.abs(output - output_hat).max(),
                1e-4
            )
            output_hat.sum().backward()
            weights_grad_sorted = torch.clone(weights_sorted.grad)
            weights_grad = torch.clone(
                weights_grad_sorted.reshape(-1, k).index_select(
                    0, q_rev_flat).view(N, H, L, k)
            )
            grad_hat = [weights_grad, torch.clone(values.grad)]
            for g1, g2 in zip(grad, grad_hat):
                self.assertLess(
                    torch.abs(g1 - g2).max(),
                    1e-3
                )

    def test_correctness(self):
        N = 12
        H = 6
        L = 1000
        S = 1000
        E = 32
        k = 32
        C = 100
        I = 10
        B = 32
        for exp in range(30):
            N = np.random.randint(1, 3)
            H = np.random.randint(1, 4)
            C = np.random.randint(10, 500)
            L = np.random.randint(C, 2000)
            E = np.random.randint(10, 128)
            S = np.random.randint(100, 1000)
            k = np.random.randint(10, 64)

            if os.getenv("VERBOSE_TESTS", ""):
                print(("Testing: N H L S E C k: "
                       "{} {} {} {} {} {} {}").format(N, H, L, S, E, C, k))

            Q = torch.randn(N, H, L, E).to(self.device)
            K = torch.randn(N, H, S, E).to(self.device)
            lengths = torch.full((N,), L, dtype=torch.int32).to(self.device)
            groups, counts = cluster_queries(Q, lengths, C, I, B)

            sorted_g, sorted_gi = torch.sort(groups.view(N*H, -1), dim=-1)
            sorted_rev_gi = torch.argsort(sorted_gi, dim=-1)
            q_offset = torch.arange(N*H, device=Q.device).unsqueeze(-1) * L
            q_flat = (sorted_gi + q_offset).reshape(-1)
            s_queries = Q.reshape(-1, E).index_select(
                0, q_flat).view(N, H, L, E)

            Q_grouped = aggregate(
                s_queries, sorted_g.view(N, H, L), 1/counts.float()
            )

            QK = torch.einsum("nhle,nhse->nhls", Q_grouped, K)
            _, topk = torch.topk(QK, k, dim=-1)
            topk = topk.contiguous()
            topk_broadcast = broadcast(
                topk.float(),
                groups,
                torch.ones_like(counts, dtype=torch.float32),
                torch.zeros((N, H, L, k), device=Q.device)
            )
            weights_sorted = torch.rand(
                N, H, L, k).to(self.device).requires_grad_(True)
            weights_sorted.retain_grad()

            q_rev_flat = (sorted_rev_gi + q_offset).reshape(-1)
            weights = torch.clone(
                weights_sorted.reshape(-1, k).index_select(
                    0, q_rev_flat).view(N, H, L, k)
            )
            weights.retain_grad()

            values = torch.randn(N, H, S, E).to(self.device).requires_grad_(True)
            self._zero_grad(weights, values)
            values_selected = values[
                torch.arange(N).view(N, 1, 1, 1).to(self.device),
                torch.arange(H).view(1, H, 1, 1).to(self.device),
                topk_broadcast.long()
            ]
            output = (weights.unsqueeze(-1)*values_selected).sum(-2)
            output.sum().backward()
            grad = [torch.clone(weights.grad), torch.clone(values.grad)]

            self._zero_grad(weights_sorted, values)
            self._zero_grad(weights, values)

            output_hat_sorted = clustered_sparse_weighted_average(
                weights_sorted, values, topk, groups, counts
            )
            output_hat = output_hat_sorted.reshape(-1, E).index_select(
                0, q_rev_flat).view(N, H, L, E)

            self.assertLess(
                torch.abs(output - output_hat).max(),
                1e-4
            )
            output_hat.sum().backward()
            weights_grad_sorted = torch.clone(weights_sorted.grad)
            weights_grad = torch.clone(
                weights_grad_sorted.reshape(-1, k).index_select(
                    0, q_rev_flat).view(N, H, L, k)
            )
            grad_hat = [weights_grad, torch.clone(values.grad)]
            for g1, g2 in zip(grad, grad_hat):
                self.assertLess(
                    torch.abs(g1 - g2).max(),
                    1e-3
                )

    def test_forward(self):
        N = 6
        H = 5
        L = 100
        S = 100
        E = 32
        C = 10
        I = 10
        B = 32
        k = 5

        for exp in range(30):
            C = np.random.randint(10, 500)
            L = np.random.randint(C, 2000)
            E = np.random.randint(10, 128)
            S = np.random.randint(100, 1000)
            k = np.random.randint(10, 64)
            if os.getenv("VERBOSE_TESTS", ""):
                print(("Testing: N H L S E C k: "
                       "{} {} {} {} {} {} {}").format(N, H, L, S, E, C, k))

            Q = torch.randn(N, H, L, E).to(self.device)
            K = torch.randn(N, H, S, E).to(self.device)
            lengths = torch.full((N,), L, dtype=torch.int32).to(self.device)
            groups, counts = cluster_queries(Q, lengths, C, I, B)

            sorted_g, sorted_gi = torch.sort(groups.view(N*H, -1), dim=-1)
            sorted_rev_gi = torch.argsort(sorted_gi, dim=-1)
            q_offset = torch.arange(N*H, device=Q.device).unsqueeze(-1) * L
            q_flat = (sorted_gi + q_offset).reshape(-1)
            s_queries = Q.reshape(-1, E).index_select(0, q_flat).view(N, H, L, E)

            Q_grouped = aggregate(
                s_queries, sorted_g.view(N, H, L), 1/counts.float()
            )

            QK = torch.einsum("nhle,nhse->nhls", Q_grouped, K)
            _, topk = torch.topk(QK, k, dim=-1)
            topk = topk.contiguous()
            topk_broadcast = broadcast(
                topk.float(),
                groups,
                torch.ones_like(counts, dtype=torch.float32),
                torch.zeros((N, H, L, k), device=Q.device)
            )

            weights_sorted = clustered_sparse_dot_product(
                s_queries, K, topk, groups, counts, lengths
            )
            weights = torch.softmax(weights_sorted, dim=-1)

            q_rev_flat = (sorted_rev_gi + q_offset).reshape(-1)
            weights = weights_sorted.reshape(-1, k).index_select(
                0, q_rev_flat).view(N, H, L, k)
            values = torch.randn(N, H, S, E).to(self.device).requires_grad_(True)
            values_selected = values[
                torch.arange(N).view(N, 1, 1, 1).to(self.device),
                torch.arange(H).view(1, H, 1, 1).to(self.device),
                topk_broadcast.long()
            ]

            output = (weights.unsqueeze(-1)*values_selected).sum(-2)
            output_hat_sorted = clustered_sparse_weighted_average(
                weights_sorted, values, topk, groups, counts
            )
            output_hat = output_hat_sorted.reshape(-1, E).index_select(
                0, q_rev_flat).view(N, H, L, E)

            self.assertLess(
                torch.abs(output_hat - output).max(),
                1e-3
            )

    @unittest.skipUnless(os.getenv("BENCHMARK_TESTS", ""), "no benchmarks")
    def test_small_forward(self):
        N = 12
        H = 8
        L = 2000
        S = 2000
        E = 32
        k = 32
        C = 100
        I = 10
        B = 32

        Q = torch.randn(N, H, L, E).to(self.device)
        K = torch.randn(N, H, S, E).to(self.device)
        lengths = torch.full((N,), L, dtype=torch.int32).to(self.device)
        groups, counts = cluster_queries(Q, lengths, C, I, B)

        sorted_g, sorted_gi = torch.sort(groups.view(N*H, -1), dim=-1)
        sorted_rev_gi = torch.argsort(sorted_gi, dim=-1)

        q_offset = torch.arange(N*H, device=Q.device).unsqueeze(-1) * L
        q_flat = (sorted_gi + q_offset).reshape(-1)
        s_queries = Q.reshape(-1, E).index_select(0, q_flat).view(N, H, L, E)
        Q_grouped = aggregate(
            s_queries, sorted_g.view(N, H, L), 1/counts.float()
        )

        QK = torch.einsum("nhle,nhse->nhls", Q_grouped, K)
        _, topk = torch.topk(QK, k, dim=-1)
        topk = topk.contiguous()
        topk_broadcast = broadcast(
            topk.float(),
            groups,
            torch.ones_like(counts, dtype=torch.float32),
            torch.zeros((N, H, L, k), device=Q.device)
        )

        weights_sorted = clustered_sparse_dot_product(
            s_queries, K, topk, groups, counts, lengths
        )
        q_rev_flat = (sorted_rev_gi + q_offset).reshape(-1)
        weights = weights_sorted.reshape(-1, k).index_select(
            0, q_rev_flat).view(N, H, L, k)

        values = torch.randn(N, H, S, E).to(self.device)
        for i in range(2000):
            output_hat = clustered_sparse_weighted_average(
                weights_sorted, values,
                topk, groups, counts
            )

        s = torch.cuda.Event(enable_timing=True)
        e = torch.cuda.Event(enable_timing=True)
        s.record()
        output_hat = clustered_sparse_weighted_average(
            weights, values,
            topk, groups, counts
        )
        e.record()
        torch.cuda.synchronize()
        t_sparse = s.elapsed_time(e)
        print('T_sparse Forward:{}'.format(t_sparse))

    @unittest.skipUnless(os.getenv("BENCHMARK_TESTS", ""), "no benchmarks")
    def test_small_backward(self):
        N = 12
        H = 8
        L = 1024
        S = 1024
        E = 64
        k = 32
        C = 100
        I = 10
        B = 32

        Q = torch.randn(N, H, L, E).to(self.device)
        K = torch.randn(N, H, S, E).to(self.device)
        lengths = torch.full((N,), L, dtype=torch.int32).to(self.device)
        groups, counts = cluster_queries(Q, lengths, C, I, B)

        sorted_g, sorted_gi = torch.sort(groups.view(N*H, -1), dim=-1)
        sorted_rev_gi = torch.argsort(sorted_gi, dim=-1)

        q_offset = torch.arange(N*H, device=Q.device).unsqueeze(-1) * L
        q_flat = (sorted_gi + q_offset).reshape(-1)
        s_queries = Q.reshape(-1, E).index_select(0, q_flat).view(N, H, L, E)
        Q_grouped = aggregate(
            s_queries, sorted_g.view(N, H, L), 1/counts.float()
        )

        QK = torch.einsum("nhle,nhse->nhls", Q_grouped, K)
        _, topk = torch.topk(QK, k, dim=-1)
        topk = topk.contiguous()
        topk_broadcast = broadcast(
            topk.float(),
            groups,
            torch.ones_like(counts, dtype=torch.float32),
            torch.zeros((N, H, L, k), device=Q.device)
        )
        weights = torch.rand(N, H, L, k).to(self.device).requires_grad_(True)
        values = torch.randn(N, H, S, E).to(self.device).requires_grad_(True)
        for i in range(2000):
            output_hat = clustered_sparse_weighted_average(
                weights, values,
                topk, groups,
                counts
            )
        self._zero_grad(weights, values)
        s = torch.cuda.Event(enable_timing=True)
        e = torch.cuda.Event(enable_timing=True)
        s.record()
        output_hat.sum().backward()
        e.record()
        torch.cuda.synchronize()
        t_sparse = s.elapsed_time(e)
        print('T_sparse Backward:{}'.format(t_sparse))


if __name__ == "__main__":
    unittest.main()



#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>,
# Apoorv Vyas <avyas@idiap.ch>
#


import unittest
import os
import numpy as np
import time

import torch

try:
    from fast_transformers.aggregate import clustered_aggregate, \
            clustered_broadcast
except ImportError:
    pass


class TestAggregateGPU(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        if not torch.cuda.is_available():
            raise unittest.SkipTest("No CUDA capable device detected")

    def test_aggregate(self):
        N = 2
        H = 4
        L = 80
        E = 2
        C = 4

        for i in range(30):
            C = np.random.randint(5, 10)
            L = np.random.randint(1, 30) * C
            E = np.random.randint(10, 128)
            if os.getenv("VERBOSE_TESTS", ""):
                print(("Testing: N H L E C: "
                       "{} {} {} {} {}").format(N, H, L, E, C))

            x = torch.rand((N, H, L, E)).cuda()
            g = (torch.arange(L) % C).view(1, 1, L).repeat(N, H, 1).int().cuda()
            f = torch.ones(N, H, C).cuda() * (C / L)
            counts = torch.ones_like(f, dtype=torch.int32) * (L // C)
            y = torch.zeros(N, H, C, E).cuda()
            lengths = torch.full((N,), L, dtype=torch.int32).to(x.device)

            sorted_g, sorted_gi = torch.sort(g.view(N*H, -1), dim=-1)
            sorted_rev_gi = torch.argsort(sorted_gi, dim=-1)

            q_offset = torch.arange(N*H, device=x.device).unsqueeze(-1) * L
            q_flat = (sorted_gi + q_offset).reshape(-1)

            # sorted queries, keys, values
            s_x = x.reshape(-1, E).index_select(0, q_flat).view(N, H, L, E)
            y = clustered_aggregate(
                s_x, sorted_g.view(N, H, -1), f, lengths, y
            )
            for i in range(C):
                self.assertLess(
                    torch.abs(
                        x[:, :, i::C, :].mean(2) - y[:, :, i, :]
                    ).max().item(),
                    1e-6
                )

    def test_aggregate_masked(self):
        N = 10
        H = 3
        L = 40
        E = 32
        C = 4

        for i in range(30):
            C = np.random.randint(5, 10)
            L = np.random.randint(2, 30) * C
            E = np.random.randint(10, 128)
            if os.getenv("VERBOSE_TESTS", ""):
                print(("Testing: N H L E C: "
                       "{} {} {} {} {}").format(N, H, L, E, C))
            x = torch.rand((N, H, L, E)).cuda()
            g = (torch.arange(L) % C).view(1, 1, L).repeat(N, H, 1).int().cuda()
            g[:, :, -C:] = C + 1
            c = (L // C) - 1

            lengths = torch.full((N,), L-C, dtype=torch.int32).to(x.device)
            f = torch.ones(N, H, C).cuda() / float(c)
            counts = torch.ones_like(f, dtype=torch.int32) * c
            y = torch.zeros(N, H, C, E).cuda()

            sorted_g, sorted_gi = torch.sort(g.view(N*H, -1), dim=-1)
            sorted_rev_gi = torch.argsort(sorted_gi, dim=-1)

            q_offset = torch.arange(N*H, device=x.device).unsqueeze(-1) * L
            q_flat = (sorted_gi + q_offset).reshape(-1)

            # sorted queries, keys, values
            s_x = x.reshape(-1, E).index_select(0, q_flat).view(N, H, L, E)
            y = clustered_aggregate(
                s_x, sorted_g.view(N, H, -1), f, lengths, y
            )

            for i in range(C):
                x_m = x[:, :, i::C, :][:, :, :-1, :].mean(2)
                self.assertLess(
                    torch.abs(
                        x_m - y[:, :, i, :]
                    ).max().item(),
                    1e-6
                )

    @unittest.skipUnless(os.getenv("BENCHMARK_TESTS", ""), "no benchmarks")
    def test_aggregate_benchmark(self):
        N = 12
        H = 8
        L = 1000
        S = 1000
        E = 64
        C = 200
        x = torch.rand(N, H, L, E).cuda()
        g = (torch.arange(L) % C).view(1, 1, L).repeat(N, H, 1).int().cuda()
        lengths = torch.full((N,), L, dtype=torch.int32).to(x.device)
        c = (0.1*torch.ones(N, H, C)).cuda()
        y = torch.zeros((N, H, C, E)).cuda()

        for i in range(2000):
            clustered_aggregate(x, g, c, lengths, y)

        s = torch.cuda.Event(enable_timing=True)
        e = torch.cuda.Event(enable_timing=True)
        s.record()
        clustered_aggregate(x, g, c, lengths, y)
        e.record()
        torch.cuda.synchronize()
        t_aggregate = s.elapsed_time(e)

        print('Aggregate Time: {}'.format(t_aggregate))

    # @unittest.skipUnless(os.getenv("BENCHMARK_TESTS", ""), "no benchmarks")
    # def test_broadcast_benchmark(self):
    #     N = 12
    #     H = 8
    #     L = 1000
    #     S = 1000
    #     E = 64
    #     C = 200

    #     y = torch.rand(N, H, C, E).cuda()
    #     g = (torch.arange(L) % C).view(1, 1, L).repeat(N, H, 1).int().cuda()
    #     c = (.1*torch.ones(N, H, C)).cuda()
    #     x = torch.zeros((N, H, L, E)).cuda()

    #     for i in range(2000):
    #         broadcast_gpu(y, g, c, x)

    #     s = torch.cuda.Event(enable_timing=True)
    #     e = torch.cuda.Event(enable_timing=True)
    #     s.record()
    #     broadcast_gpu(y, g, c, x)
    #     e.record()
    #     torch.cuda.synchronize()
    #     t_broadcast = s.elapsed_time(e)

    #     print('Broadcast Time: {}'.format(t_broadcast))


if __name__ == "__main__":
    unittest.main()



#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>
#

"""Boilerplate code for dealing with fast_transformers modules."""


def make_mirror(src_module, dst_module):
    """Sets the parameters of src_module to dst_module so that they share the
    same parameters.

    Most noteable usecase is to make a recurrent transformer mirror of a batch
    transformer for fast inference.

    Arguments
    ---------
        src_module: Module to take the parameters from
        dst_module: Module to set the parameters to

    Returns
    -------
        None, it changes dst_module in place
    """
    def setattr_recursive(mod, key, value):
        key, *next_key = key.split(".", maxsplit=1)
        if not next_key:
            setattr(mod, key, value)
        else:
            setattr_recursive(getattr(mod, key), next_key[0], value)

    for name, param in src_module.named_parameters():
        setattr_recursive(dst_module, name, param)



#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>
#

import torch

from .local_product_cpu import local_dot_product as local_dot_product_cpu, \
    local_dot_backward as local_dot_backward_cpu, \
    local_weighted_average as local_weighted_average_cpu, \
    local_weighted_average_backward as local_weighted_average_backward_cpu

try:
    from .local_product_cuda import \
        local_dot_product as local_dot_product_cuda, \
        local_dot_backward as local_dot_backward_cuda, \
        local_weighted_average as local_weighted_average_cuda, \
        local_weighted_average_backward as local_weighted_average_backward_cuda
except ImportError:
    local_dot_product_cuda = None
    local_dot_backward_cuda = None
    local_weighted_average_cuda = None
    local_weighted_average_backward_cuda = None


class LocalDotProduct(torch.autograd.Function):
    """Compute the dot product of the queries and keys but only consider a
    local neighborhood of each query."""
    dot = {
        "cpu": local_dot_product_cpu,
        "cuda": local_dot_product_cuda
    }
    dot_backward = {
        "cpu": local_dot_backward_cpu,
        "cuda": local_dot_backward_cuda
    }

    @staticmethod
    def forward(ctx, queries, keys, attn_mask, key_lengths, local_context):
        # Save the inputs for the gradient computation
        ctx.save_for_backward(queries, keys, key_lengths)
        ctx.local_context = local_context

        return LocalDotProduct.dot[queries.device.type](
            queries,
            keys,
            attn_mask,
            key_lengths,
            local_context
        )

    @staticmethod
    def backward(ctx, grad_input):
        queries, keys, key_lengths = ctx.saved_tensors
        local_context = ctx.local_context

        grads = LocalDotProduct.dot_backward[queries.device.type](
            queries,
            keys,
            key_lengths,
            grad_input,
            local_context
        )

        # plus 3 None for masks and local_context
        return grads + (None, None, None)


class LocalWeightedAverage(torch.autograd.Function):
    """Compute the weighted average of the values with the local attention."""
    avg = {
        "cpu": local_weighted_average_cpu,
        "cuda": local_weighted_average_cuda
    }
    avg_backward = {
        "cpu": local_weighted_average_backward_cpu,
        "cuda": local_weighted_average_backward_cuda
    }

    @staticmethod
    def forward(ctx, A, V):
        # Save the inputs for the gradient computation
        ctx.save_for_backward(A, V)

        return LocalWeightedAverage.avg[A.device.type](A, V)

    @staticmethod
    def backward(ctx, grad_input):
        A, V = ctx.saved_tensors
        return LocalWeightedAverage.avg_backward[A.device.type](
            A, V, grad_input
        )


# Alias the autograd functions to python style snake case naming
local_dot_product = LocalDotProduct.apply
local_weighted_average = LocalWeightedAverage.apply



#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>
#


class Event(object):
    """The Event is the base class for all events that are dispatched from any
    transformer module.

    This class defines only the basic attributes of an event without any
    payload.

    Arguments
    ---------
        source: torch.nn.Module instance that dispatched this event
    """
    def __init__(self, source):
        self.source = source


class AttentionEvent(Event):
    """An event containing an attention matrix.

    Arguments
    ---------
        source: torch.nn.Module instance that dispatched this event
        attention_matrix: torch.tensor of the multihead attention matrix
                          computed in the corresponding attention layer
    """
    def __init__(self, source, attention_matrix):
        super(AttentionEvent, self).__init__(source)
        self.attention_matrix = attention_matrix


class QKVEvent(Event):
    """An event containing the queries, keys and values projected in their
    multiple heads.

    Arguments
    ---------
        source: torch.nn.Module instance that dispatched this event
        queries: torch.tensor containing the queries in shape NLHE
        keys: torch.tensor containing the keys in shape NSHE
        values: torch.tensor containing the values in shape NSHD
    """
    def __init__(self, source, queries, keys, values):
        super(QKVEvent, self).__init__(source)
        self.queries = queries
        self.keys = keys
        self.values = values
